@@ -45,6 +45,10 @@
 #define READ_CHUNK_SIZE 8192
 #define MIME_READ_CHUNK_SIZE 1024
 #define RECURSE_LEVEL_MAX 4
+#define DIR_MIME_TYPE "x-directory/normal"
+#define BLOCK_DEVICE_TYPE "x-special/device-block"
+#define EXTINF "#EXTINF:"
+#define DEBUG(x) { if (parser->priv->debug != FALSE) x; }
 
 typedef TotemPlParserResult (*PlaylistCallback) (TotemPlParser *parser, const char *url, gpointer data);
 static gboolean totem_pl_parser_scheme_is_ignored (TotemPlParser *parser, const char *url);
@@ -52,6 +56,15 @@ static gboolean totem_pl_parser_ignore (TotemPlParser *parser, const char *url);
 static TotemPlParserResult totem_pl_parser_parse_internal (TotemPlParser *parser, const char *url);
 static TotemPlParserResult totem_pl_parser_add_asx (TotemPlParser *parser, const char *url, gpointer data);
 
+static void totem_pl_parser_set_property (GObject *object,
+					  guint prop_id,
+					  const GValue *value,
+					  GParamSpec *pspec);
+static void totem_pl_parser_get_property (GObject *object,
+					  guint prop_id,
+					  GValue *value,
+					  GParamSpec *pspec);
+
 typedef struct {
 	char *mimetype;
 	PlaylistCallback func;
@@ -63,6 +76,14 @@ struct TotemPlParserPrivate
 	GList *ignore_mimetypes;
 	guint recurse_level;
 	gboolean fallback;
+	gboolean recurse;
+	gboolean debug;
+};
+
+enum {
+	PROP_NONE,
+	PROP_RECURSE,
+	PROP_DEBUG
 };
 
 /* Signals */
@@ -91,6 +112,25 @@ totem_pl_parser_class_init (TotemPlParserClass *klass)
 	parent_class = g_type_class_peek_parent (klass);
 
 	object_class->finalize = totem_pl_parser_finalize;
+	object_class->set_property = totem_pl_parser_set_property;
+	object_class->get_property = totem_pl_parser_get_property;
+
+	/* properties */
+	g_object_class_install_property (object_class,
+					 PROP_RECURSE,
+					 g_param_spec_boolean ("recurse",
+							       "recurse",
+							       "Whether or not to process URLs further", 
+							       TRUE,
+							       G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
+
+	g_object_class_install_property (object_class,
+					 PROP_DEBUG,
+					 g_param_spec_boolean ("debug",
+							       "debug",
+							       "Whether or not to enable debugging output", 
+							       FALSE,
+							       G_PARAM_READWRITE));
 
 	/* Signals */
 	totem_pl_parser_table_signals[ENTRY] =
@@ -119,6 +159,50 @@ totem_pl_parser_class_init (TotemPlParserClass *klass)
 			      G_TYPE_NONE, 1, G_TYPE_STRING);
 }
 
+static void
+totem_pl_parser_set_property (GObject *object,
+			      guint prop_id,
+			      const GValue *value,
+			      GParamSpec *pspec)
+{
+	TotemPlParser *parser = TOTEM_PL_PARSER (object);
+
+	switch (prop_id)
+	{
+	case PROP_RECURSE:
+		parser->priv->recurse = g_value_get_boolean (value);
+		break;
+	case PROP_DEBUG:
+		parser->priv->debug = g_value_get_boolean (value);
+		break;
+	default:
+		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+		break;
+	}
+}
+
+static void
+totem_pl_parser_get_property (GObject *object,
+			      guint prop_id,
+			      GValue *value,
+			      GParamSpec *pspec)
+{
+	TotemPlParser *parser = TOTEM_PL_PARSER (object);
+
+	switch (prop_id)
+	{
+	case PROP_RECURSE:
+		g_value_set_boolean (value, parser->priv->recurse);
+		break;
+	case PROP_DEBUG:
+		g_value_set_boolean (value, parser->priv->debug);
+		break;
+	default:
+		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+		break;
+	}
+}
+
 GQuark
 totem_pl_parser_error_quark (void)
 {
@@ -136,7 +220,7 @@ totem_pl_parser_new (void)
 }
 
 static const char *
-my_gnome_vfs_get_mime_type_with_data (const char *uri, gpointer *data)
+my_gnome_vfs_get_mime_type_with_data (const char *uri, gpointer *data, TotemPlParser *parser)
 {
 	GnomeVFSResult result;
 	GnomeVFSHandle *handle;
@@ -147,10 +231,25 @@ my_gnome_vfs_get_mime_type_with_data (const char *uri, gpointer *data)
 
 	*data = NULL;
 
+	/* Stat for a block device, we're screwed as far as speed
+	 * is concerned now */
+	if (g_str_has_prefix (uri, "file://") != FALSE) {
+		struct stat buf;
+		if (stat (uri + strlen ("file://"), &buf) == 0) {
+			if (S_ISBLK (buf.st_mode))
+				return BLOCK_DEVICE_TYPE;
+		}
+	}
+
 	/* Open the file. */
 	result = gnome_vfs_open (&handle, uri, GNOME_VFS_OPEN_READ);
-	if (result != GNOME_VFS_OK)
+	if (result != GNOME_VFS_OK) {
+		if (result == GNOME_VFS_ERROR_IS_DIRECTORY)
+			return DIR_MIME_TYPE;
+		DEBUG(g_print ("URL '%s' couldn't be opened in _get_mime_type_with_data: '%s'\n", uri, gnome_vfs_result_to_string (result)));
 		return NULL;
+	}
+	DEBUG(g_print ("URL '%s' was opened successfully in _get_mime_type_with_data:\n", uri));
 
 	/* Read the whole thing, up to MIME_READ_CHUNK_SIZE */
 	buffer = NULL;
@@ -179,9 +278,14 @@ my_gnome_vfs_get_mime_type_with_data (const char *uri, gpointer *data)
 	} while (result == GNOME_VFS_OK
 			&& total_bytes_read < MIME_READ_CHUNK_SIZE);
 
-	/* Close the file. */
-	result = gnome_vfs_close (handle);
+	/* Close the file but don't overwrite the possible error */
+	if (result != GNOME_VFS_OK && result != GNOME_VFS_ERROR_EOF)
+		gnome_vfs_close (handle);
+	else
+		result = gnome_vfs_close (handle);
+
 	if (result != GNOME_VFS_OK) {
+		DEBUG(g_print ("URL '%s' couldn't be read or closed in _get_mime_type_with_data: '%s'\n", uri, gnome_vfs_result_to_string (result)));
 		g_free (buffer);
 		return NULL;
 	}
@@ -196,14 +300,18 @@ my_gnome_vfs_get_mime_type_with_data (const char *uri, gpointer *data)
 	return mimetype;
 }
 
-static char*
-totem_pl_parser_set_method (const char *url, const char *method)
+static char *
+totem_pl_parser_read_entire_file (const char *url, int *size)
 {
-	char *start;
+	char *contents;
 
-	g_return_val_if_fail (strstr (url, "://"), NULL);
-	start = strstr (url, "://");
-	return g_strdup_printf ("%s%s", method, start);
+	if (gnome_vfs_read_entire_file (url, size, &contents) != GNOME_VFS_OK)
+		return NULL;
+
+	contents = g_realloc ((gpointer) contents, *size + 1);
+	contents[*size] = '\0';
+
+	return contents;
 }
 
 static char*
@@ -266,12 +374,13 @@ totem_pl_parser_num_entries (TotemPlParser *parser, GtkTreeModel *model,
 	{
 		GtkTreeIter iter;
 		char *path, *url, *title;
+		gboolean custom_title;
 
 		path = g_strdup_printf ("%d", i - 1);
 		gtk_tree_model_get_iter_from_string (model, &iter, path);
 		g_free (path);
 
-		func (model, &iter, &url, &title, user_data);
+		func (model, &iter, &url, &title, &custom_title, user_data);
 		if (totem_pl_parser_scheme_is_ignored (parser, url) != FALSE)
 			ignored++;
 
@@ -385,12 +494,13 @@ totem_pl_parser_write_pls (TotemPlParser *parser, GtkTreeModel *model,
 	for (i = 1; i <= num_entries_total; i++) {
 		GtkTreeIter iter;
 		char *path, *url, *title, *relative;
+		gboolean custom_title;
 
 		path = g_strdup_printf ("%d", i - 1);
 		gtk_tree_model_get_iter_from_string (model, &iter, path);
 		g_free (path);
 
-		func (model, &iter, &url, &title, user_data);
+		func (model, &iter, &url, &title, &custom_title, user_data);
 
 		if (totem_pl_parser_scheme_is_ignored (parser, url) != FALSE)
 		{
@@ -413,6 +523,11 @@ totem_pl_parser_write_pls (TotemPlParser *parser, GtkTreeModel *model,
 			return FALSE;
 		}
 
+		if (custom_title == FALSE) {
+			g_free (title);
+			continue;
+		}
+
 		buf = g_strdup_printf ("Title%d=%s\n", i, title);
 		success = write_string (handle, buf, error);
 		g_free (buf);
@@ -471,6 +586,7 @@ totem_pl_parser_write_m3u (TotemPlParser *parser, GtkTreeModel *model,
 	int num_entries_total, i;
 	gboolean success;
 	char *buf;
+	char *cr;
 
 	res = gnome_vfs_open (&handle, output, GNOME_VFS_OPEN_WRITE);
 	if (res == GNOME_VFS_ERROR_NOT_FOUND) {
@@ -490,17 +606,21 @@ totem_pl_parser_write_m3u (TotemPlParser *parser, GtkTreeModel *model,
 		return FALSE;
 	}
 
+	cr = dos_compatible ? "\r\n" : "\n";
 	num_entries_total = gtk_tree_model_iter_n_children (model, NULL);
+	if (num_entries_total == 0)
+		return TRUE;
 
 	for (i = 1; i <= num_entries_total; i++) {
 		GtkTreeIter iter;
-		char *path, *url, *title;
+		char *path, *url, *title, *path2;
+		gboolean custom_title;
 
 		path = g_strdup_printf ("%d", i - 1);
 		gtk_tree_model_get_iter_from_string (model, &iter, path);
 		g_free (path);
 
-		func (model, &iter, &url, &title, user_data);
+		func (model, &iter, &url, &title, &custom_title, user_data);
 
 		if (totem_pl_parser_scheme_is_ignored (parser, url) != FALSE)
 		{
@@ -509,22 +629,37 @@ totem_pl_parser_write_m3u (TotemPlParser *parser, GtkTreeModel *model,
 			continue;
 		}
 
-		if (dos_compatible != FALSE)
-		{
-			char *dos;
+		if (custom_title != FALSE) {
+			buf = g_strdup_printf (EXTINF",%s%s", title, cr);
+			success = write_string (handle, buf, error);
+			g_free (buf);
+			if (success == FALSE) {
+				g_free (title);
+				g_free (url);
+				gnome_vfs_close (handle);
+				return FALSE;
+			}
+		}
+		g_free (title);
 
-			dos = totem_pl_parser_url_to_dos (url, output);
-			buf = g_strdup_printf ("%s\r\n", dos);
-			g_free (dos);
+		if (dos_compatible == FALSE) {
+			char *tmp;
+			tmp = totem_pl_parser_relative (url, output);
+			if (tmp == NULL && g_str_has_prefix (url, "file:")) {
+				path2 = g_filename_from_uri (url, NULL, NULL);
+			} else {
+				path2 = tmp;
+			}
 		} else {
-			char *relative;
-
-			relative = totem_pl_parser_relative (url, output);
-			buf = g_strdup_printf ("%s\n", relative);
-			g_free (relative);
+			path2 = totem_pl_parser_url_to_dos (url, output);
 		}
 
-		success = write_string (handle, url, error);
+		buf = g_strdup_printf ("%s%s", path2 ? path2 : url, cr);
+		g_free (path2);
+		g_free (url);
+
+		success = write_string (handle, buf, error);
+		g_free (buf);
 
 		if (success == FALSE)
 		{
@@ -533,6 +668,8 @@ totem_pl_parser_write_m3u (TotemPlParser *parser, GtkTreeModel *model,
 		}
 	}
 
+	gnome_vfs_close (handle);
+
 	return TRUE;
 }
 
@@ -581,10 +718,15 @@ read_ini_line_int (char **lines, const char *key)
 		return -1;
 
 	for (i = 0; (lines[i] != NULL && retval == -1); i++) {
-		if (g_ascii_strncasecmp (lines[i], key, strlen (key)) == 0) {
+		char *line = lines[i];
+
+		while (*line == '\t' || *line == ' ')
+			line++;
+
+		if (g_ascii_strncasecmp (line, key, strlen (key)) == 0) {
 			char **bits;
 
-			bits = g_strsplit (lines[i], "=", 2);
+			bits = g_strsplit (line, "=", 2);
 			if (bits[0] == NULL || bits [1] == NULL) {
 				g_strfreev (bits);
 				return -1;
@@ -608,11 +750,16 @@ read_ini_line_string (char **lines, const char *key, gboolean dos_mode)
 		return NULL;
 
 	for (i = 0; (lines[i] != NULL && retval == NULL); i++) {
-		if (g_ascii_strncasecmp (lines[i], key, strlen (key)) == 0) {
+		char *line = lines[i];
+
+		while (*line == '\t' || *line == ' ')
+			line++;
+
+		if (g_ascii_strncasecmp (line, key, strlen (key)) == 0) {
 			char **bits;
 			ssize_t len;
 
-			bits = g_strsplit (lines[i], "=", 2);
+			bits = g_strsplit (line, "=", 2);
 			if (bits[0] == NULL || bits [1] == NULL) {
 				g_strfreev (bits);
 				return NULL;
@@ -693,7 +840,8 @@ totem_pl_parser_add_ram (TotemPlParser *parser, const char *url, gpointer data)
 	int size, i;
 	const char *split_char;
 
-	if (gnome_vfs_read_entire_file (url, &size, &contents) != GNOME_VFS_OK)
+	contents = totem_pl_parser_read_entire_file (url, &size);
+	if (contents == NULL)
 		return TOTEM_PL_PARSER_RESULT_ERROR;
 
 	/* figure out whether we're a unix or dos RAM file */
@@ -752,7 +900,8 @@ totem_pl_parser_add_asf_reference_parser (TotemPlParser *parser,
 	char *contents, **lines, *ref, *split_char;
 	int size;
 
-	if (gnome_vfs_read_entire_file (url, &size, &contents) != GNOME_VFS_OK)
+	contents = totem_pl_parser_read_entire_file (url, &size);
+	if (contents == NULL)
 		return TOTEM_PL_PARSER_RESULT_ERROR;
 
 	if (strstr(contents,"\x0d") == NULL) {
@@ -796,7 +945,8 @@ totem_pl_parser_add_asf_parser (TotemPlParser *parser,
 		return totem_pl_parser_add_asf_reference_parser (parser, url, data);
 	}
 
-	if (gnome_vfs_read_entire_file (url, &size, &contents) != GNOME_VFS_OK)
+	contents = totem_pl_parser_read_entire_file (url, &size);
+	if (contents == NULL)
 		return TOTEM_PL_PARSER_RESULT_ERROR;
 
 	if (size <= 4) {
@@ -824,6 +974,7 @@ totem_pl_parser_add_pls_with_contents (TotemPlParser *parser, const char *url, c
 	int i, num_entries;
 	char *split_char, *playlist_title;
 	gboolean dos_mode = FALSE;
+	gboolean fallback;
 
 	/* figure out whether we're a unix pls or dos pls */
 	if (strstr(contents,"\x0d") == NULL) {
@@ -844,8 +995,10 @@ totem_pl_parser_add_pls_with_contents (TotemPlParser *parser, const char *url, c
 
 	if (lines[i] == NULL
 			|| g_ascii_strncasecmp (lines[i], "[playlist]",
-				(gsize)strlen ("[playlist]")) != 0)
+				(gsize)strlen ("[playlist]")) != 0) {
+		DEBUG (g_print ("Couldn't recognise URL '%s' as a .pls playlist\n", url));
 		goto bail;
+	}
 
 	playlist_title = read_ini_line_string (lines,
 			"X-GNOME-Title", dos_mode);
@@ -858,8 +1011,10 @@ totem_pl_parser_add_pls_with_contents (TotemPlParser *parser, const char *url, c
 
 	/* numberofentries=? */
 	num_entries = read_ini_line_int (lines, "numberofentries");
-	if (num_entries == -1)
+	if (num_entries == -1) {
+		DEBUG (g_print ("Couldn't get the numberofentries for '%s'\n", url));
 		goto bail;
+	}
 
 	retval = TOTEM_PL_PARSER_RESULT_SUCCESS;
 
@@ -888,10 +1043,14 @@ totem_pl_parser_add_pls_with_contents (TotemPlParser *parser, const char *url, c
 			continue;
 		}
 
-		if (strstr (file, "://") != NULL
-				|| file[0] == G_DIR_SEPARATOR) {
-			totem_pl_parser_add_one_url_ext (parser,
-					file, title, genre);
+		fallback = parser->priv->fallback;
+		if (parser->priv->recurse)
+			parser->priv->fallback = FALSE;
+
+		if (strstr (file, "://") != NULL || file[0] == G_DIR_SEPARATOR) {
+			if (totem_pl_parser_parse_internal (parser, file) != TOTEM_PL_PARSER_RESULT_SUCCESS) {
+				totem_pl_parser_add_one_url_ext (parser, file, title, genre);
+			}
 		} else {
 			char *uri, *base, *escaped;
 
@@ -901,14 +1060,16 @@ totem_pl_parser_add_pls_with_contents (TotemPlParser *parser, const char *url, c
 
 			uri = g_strdup_printf ("%s/%s", base, escaped);
 
-			totem_pl_parser_add_one_url_ext (parser,
-					uri, title, genre);
+			if (totem_pl_parser_parse_internal (parser, uri) != TOTEM_PL_PARSER_RESULT_SUCCESS) {
+				totem_pl_parser_add_one_url_ext (parser, uri, title, genre);
+			}
 
 			g_free (escaped);
 			g_free (uri);
 			g_free (base);
 		}
 
+		parser->priv->fallback = fallback;
 		g_free (file);
 		g_free (title);
 		g_free (genre);
@@ -934,7 +1095,8 @@ totem_pl_parser_add_pls (TotemPlParser *parser, const char *url, gpointer data)
 	char *contents;
 	int size;
 
-	if (gnome_vfs_read_entire_file (url, &size, &contents) != GNOME_VFS_OK)
+	contents = totem_pl_parser_read_entire_file (url, &size);
+	if (contents == NULL)
 		return TOTEM_PL_PARSER_RESULT_ERROR;
 
 	if (size == 0) {
@@ -948,25 +1110,30 @@ totem_pl_parser_add_pls (TotemPlParser *parser, const char *url, gpointer data)
 	return retval;
 }
 
+
 static const char *
 totem_pl_parser_get_extinfo_title (gboolean extinfo, char **lines, int i)
 {
-	const char *retval;
+	const char *extinf, *comma;
 
-	if (extinfo == FALSE || lines == NULL)
+	if (extinfo == FALSE || lines == NULL || i <= 0)
 		return NULL;
 
-	if (i == 0)
+	/* It's bound to have an EXTINF if we have extinfo */
+	extinf = lines[i-1] + strlen(EXTINF);
+	if (extinf[0] == '\0')
 		return NULL;
+	comma = strstr (extinf, ",");
 
-	retval = strstr (lines[i-1], "#EXTINF:");
-	retval = strstr (retval, ",");
-	if (retval == NULL || retval[0] == '\0')
-		return NULL;
+	if (comma == NULL || comma[1] == '\0') {
+		if (extinf[1] == '\0')
+			return NULL;
+		return extinf;
+	}
 
-	retval++;
+	comma++;
 
-	return retval;
+	return comma;
 }
 
 static TotemPlParserResult
@@ -978,7 +1145,8 @@ totem_pl_parser_add_m3u (TotemPlParser *parser, const char *url, gpointer data)
 	const char *split_char;
 	gboolean extinfo;
 
-	if (gnome_vfs_read_entire_file (url, &size, &contents) != GNOME_VFS_OK)
+	contents = totem_pl_parser_read_entire_file (url, &size);
+	if (contents == NULL)
 		return TOTEM_PL_PARSER_RESULT_ERROR;
 
 	/* .pls files with a .m3u extension, the nasties */
@@ -1010,8 +1178,7 @@ totem_pl_parser_add_m3u (TotemPlParser *parser, const char *url, gpointer data)
 
 		/* Ignore comments, but mark it if we have extra info */
 		if (lines[i][0] == '#') {
-			if (strstr (lines[i], "#EXTINF") != NULL)
-				extinfo = TRUE;
+			extinfo = g_str_has_prefix (lines[i], EXTINF);
 			continue;
 		}
 
@@ -1074,12 +1241,23 @@ parse_asx_entry (TotemPlParser *parser, char *base, xmlDocPtr doc,
 		if (node->name == NULL)
 			continue;
 
-		/* ENTRY should only have one ref and one title nodes */
+		/* ENTRY can only have one title node but multiple REFs */
 		if (g_ascii_strcasecmp ((char *)node->name, "ref") == 0
 				|| g_ascii_strcasecmp ((char *)node->name, "entryref") == 0) {
-			url = xmlGetProp (node, (guchar *)"href");
-			if (url == NULL)
-				url = xmlGetProp (node, (guchar *)"HREF");
+			unsigned char *tmp;
+
+			tmp = xmlGetProp (node, (guchar *)"href");
+			if (tmp == NULL)
+				tmp = xmlGetProp (node, (guchar *)"HREF");
+			if (tmp == NULL)
+				continue;
+			if (url == NULL || g_str_has_prefix ((char *)tmp, "mms:") != FALSE) {
+				g_free (url);
+				url = tmp;
+			} else {
+				g_free (tmp);
+			}
+
 			continue;
 		}
 
@@ -1100,13 +1278,6 @@ parse_asx_entry (TotemPlParser *parser, char *base, xmlDocPtr doc,
 
 	g_free (url);
 
-	if (g_str_has_prefix (fullpath, "mms:") != FALSE) {
-		char *tmp;
-		tmp = totem_pl_parser_set_method (fullpath, "mmsh");
-		g_free (fullpath);
-		fullpath = tmp;
-	}
-
 	/* .asx files can contain references to other .asx files */
 	retval = totem_pl_parser_parse_internal (parser, fullpath);
 	if (retval != TOTEM_PL_PARSER_RESULT_SUCCESS) {
@@ -1168,7 +1339,8 @@ totem_pl_parser_add_asx (TotemPlParser *parser, const char *url, gpointer data)
 	int size;
 	gboolean retval = TOTEM_PL_PARSER_RESULT_UNHANDLED;
 
-	if (gnome_vfs_read_entire_file (url, &size, &contents) != GNOME_VFS_OK)
+	contents = totem_pl_parser_read_entire_file (url, &size);
+	if (contents == NULL)
 		return FALSE;
 
 	doc = xmlParseMemory (contents, size);
@@ -1194,6 +1366,26 @@ totem_pl_parser_add_asx (TotemPlParser *parser, const char *url, gpointer data)
 	return retval;
 }
 
+static TotemPlParserResult
+totem_pl_parser_add_block (TotemPlParser *parser, const char *url, gpointer data)
+{
+	MediaType type;
+	char *media_url;
+	GError *err = NULL;
+
+	type = totem_cd_detect_type_with_url (url, &media_url, &err);
+	if (err != NULL)
+		DEBUG(g_print ("Couldn't get CD type for URL '%s': %s\n", url, err->message));
+	if (type == MEDIA_TYPE_DATA || media_url == NULL)
+		return TOTEM_PL_PARSER_RESULT_UNHANDLED;
+	else if (type == MEDIA_TYPE_ERROR)
+		return TOTEM_PL_PARSER_RESULT_ERROR;
+
+	totem_pl_parser_add_one_url (parser, media_url, NULL);
+	g_free (media_url);
+	return TOTEM_PL_PARSER_RESULT_SUCCESS;
+}
+
 static TotemPlParserResult
 totem_pl_parser_add_ra (TotemPlParser *parser, const char *url, gpointer data)
 {
@@ -1243,7 +1435,7 @@ parse_smil_entry (TotemPlParser *parser, char *base, xmlDocPtr doc,
 			continue;
 
 		/* ENTRY should only have one ref and one title nodes */
-		if (g_ascii_strcasecmp ((char *)node->name, "video") == 0) {
+		if (g_ascii_strcasecmp ((char *)node->name, "video") == 0 || g_ascii_strcasecmp ((char *)node->name, "audio") == 0) {
 			url = xmlGetProp (node, (guchar *)"src");
 			title = xmlGetProp (node, (guchar *)"title");
 
@@ -1296,7 +1488,8 @@ totem_pl_parser_add_smil (TotemPlParser *parser, const char *url, gpointer data)
 	int size;
 	TotemPlParserResult retval = TOTEM_PL_PARSER_RESULT_UNHANDLED;
 
-	if (gnome_vfs_read_entire_file (url, &size, &contents) != GNOME_VFS_OK)
+	contents = totem_pl_parser_read_entire_file (url, &size);
+	if (contents == NULL)
 		return TOTEM_PL_PARSER_RESULT_ERROR;
 
 	doc = xmlParseMemory (contents, size);
@@ -1329,11 +1522,15 @@ totem_pl_parser_add_smil (TotemPlParser *parser, const char *url, gpointer data)
 static TotemPlParserResult
 totem_pl_parser_add_asf (TotemPlParser *parser, const char *url, gpointer data)
 {
-	if (data != NULL &&
-		(g_str_has_prefix (data, "[Reference]") == FALSE
+	if (data == NULL) {
+		totem_pl_parser_add_one_url (parser, url, NULL);
+		return TOTEM_PL_PARSER_RESULT_SUCCESS;
+	}
+
+	if (g_str_has_prefix (data, "[Reference]") == FALSE
 		 && g_ascii_strncasecmp (data, "<ASX", strlen ("<ASX")) != 0
 		 && strstr (data, "<ASX") == NULL
-		 && strstr (data, "<asx") == NULL)
+		 && strstr (data, "<asx") == NULL
 		 && g_str_has_prefix (data, "ASF ") == FALSE) {
 		totem_pl_parser_add_one_url (parser, url, NULL);
 		return TOTEM_PL_PARSER_RESULT_SUCCESS;
@@ -1342,6 +1539,70 @@ totem_pl_parser_add_asf (TotemPlParser *parser, const char *url, gpointer data)
 	return totem_pl_parser_add_asf_parser (parser, url, data);
 }
 
+static TotemPlParserResult
+totem_pl_parser_add_quicktime_metalink (TotemPlParser *parser, const char *url, gpointer data)
+{
+	xmlDocPtr doc;
+	xmlNodePtr node;
+	char *contents = NULL, *src;
+	int size;
+
+	contents = totem_pl_parser_read_entire_file (url, &size);
+	if (contents == NULL)
+		return TOTEM_PL_PARSER_RESULT_ERROR;
+
+	doc = xmlParseMemory (contents, size);
+	if (doc == NULL)
+		doc = xmlRecoverMemory (contents, size);
+	g_free (contents);
+
+	/* If the document has no root, or no name */
+	if(!doc || !doc->children
+			|| !doc->children->name
+			|| g_ascii_strcasecmp ((char *)doc->children->name,
+				"quicktime") != 0) {
+		if (doc != NULL)
+			xmlFreeDoc (doc);
+		return TOTEM_PL_PARSER_RESULT_ERROR;
+	}
+
+	if (strstr ((char *) doc->children->content, "type=\"application/x-quicktime-media-link\"") == NULL) {
+		xmlFreeDoc (doc);
+		return TOTEM_PL_PARSER_RESULT_ERROR;
+	}
+
+	node = doc->children->next;
+	if (!node || !node->name
+			|| g_ascii_strcasecmp ((char *) node->name,
+				"embed") != 0) {
+		xmlFreeDoc (doc);
+		return TOTEM_PL_PARSER_RESULT_ERROR;
+	}
+
+	src = (char *) xmlGetProp (node, (guchar *)"src");
+	if (!src) {
+		xmlFreeDoc (doc);
+		return TOTEM_PL_PARSER_RESULT_ERROR;
+	}
+
+	totem_pl_parser_add_one_url (parser, src, NULL);
+
+	xmlFreeDoc (doc);
+
+	return TOTEM_PL_PARSER_RESULT_SUCCESS;
+}
+
+static TotemPlParserResult
+totem_pl_parser_add_quicktime (TotemPlParser *parser, const char *url, gpointer data)
+{
+	if (data == NULL || strstr (data, "<?quicktime") == NULL) {
+		totem_pl_parser_add_one_url (parser, url, NULL);
+		return TOTEM_PL_PARSER_RESULT_SUCCESS;
+	}
+
+	return totem_pl_parser_add_quicktime_metalink (parser, url, data);
+}
+
 static TotemPlParserResult
 totem_pl_parser_add_desktop (TotemPlParser *parser, const char *url, gpointer data)
 {
@@ -1349,7 +1610,8 @@ totem_pl_parser_add_desktop (TotemPlParser *parser, const char *url, gpointer da
 	const char *path, *display_name, *type;
 	int size;
 
-	if (gnome_vfs_read_entire_file (url, &size, &contents) != GNOME_VFS_OK)
+	contents = totem_pl_parser_read_entire_file (url, &size);
+	if (contents == NULL)
 		return TOTEM_PL_PARSER_RESULT_ERROR;
 
 	lines = g_strsplit (contents, "\n", 0);
@@ -1504,6 +1766,7 @@ static PlaylistTypes special_types[] = {
 	{ "audio/x-ms-asx", totem_pl_parser_add_asx },
 	{ "audio/x-scpls", totem_pl_parser_add_pls },
 	{ "application/x-smil", totem_pl_parser_add_smil },
+	{ "application/smil", totem_pl_parser_add_smil },
 	{ "application/x-gnome-app-info", totem_pl_parser_add_desktop },
 	{ "application/x-desktop", totem_pl_parser_add_desktop },
 	{ "x-directory/normal", totem_pl_parser_add_directory },
@@ -1511,6 +1774,7 @@ static PlaylistTypes special_types[] = {
 	{ "audio/x-ms-wax", totem_pl_parser_add_asx },
 	{ "application/x-cd-image", totem_pl_parser_add_iso },
 	{ "application/x-cue", totem_pl_parser_add_cue },
+	{ "x-special/device-block", totem_pl_parser_add_block },
 };
 
 static PlaylistTypes ignore_types[] = {
@@ -1527,9 +1791,11 @@ static PlaylistTypes dual_types[] = {
 	{ "audio/x-pn-realaudio", totem_pl_parser_add_ra },
 	{ "application/vnd.rn-realmedia", totem_pl_parser_add_ra },
 	{ "audio/x-pn-realaudio-plugin", totem_pl_parser_add_ra },
+	{ "audio/vnd.rn-realaudio", totem_pl_parser_add_ra },
 	{ "text/plain", totem_pl_parser_add_ra },
 	{ "video/x-ms-asf", totem_pl_parser_add_asf },
 	{ "video/x-ms-wmv", totem_pl_parser_add_asf },
+	{ "video/quicktime", totem_pl_parser_add_quicktime },
 };
 
 static gboolean
@@ -1607,15 +1873,30 @@ totem_pl_parser_parse_internal (TotemPlParser *parser, const char *url)
 	const char *mimetype;
 	guint i;
 	gpointer data = NULL;
-	TotemPlParserResult ret = FALSE;
+	TotemPlParserResult ret = TOTEM_PL_PARSER_RESULT_ERROR;
 	char *super;
+	gboolean found = FALSE;
 
 	if (parser->priv->recurse_level > RECURSE_LEVEL_MAX)
 		return TOTEM_PL_PARSER_RESULT_ERROR;
 
+	/* Shouldn't gnome-vfs have a list of schemes it supports? */
+	if (g_str_has_prefix (url, "mms") != FALSE
+			|| g_str_has_prefix (url, "rtsp") != FALSE) {
+		DEBUG(g_print ("URL '%s' is MMS or RTSP, ignoring\n", url));
+		totem_pl_parser_add_one_url (parser, url, NULL);
+		return TOTEM_PL_PARSER_RESULT_SUCCESS;
+	}
+
+	if (!parser->priv->recurse && parser->priv->recurse_level > 0) {
+		return TOTEM_PL_PARSER_RESULT_UNHANDLED;
+	}
+
 	mimetype = gnome_vfs_mime_type_from_name (url);
+	DEBUG(g_print ("_mime_type_from_name for '%s' returned '%s'\n", url, mimetype));
 	if (mimetype == NULL || strcmp (GNOME_VFS_MIME_TYPE_UNKNOWN, mimetype) == 0) {
-		mimetype = my_gnome_vfs_get_mime_type_with_data (url, &data);
+		mimetype = my_gnome_vfs_get_mime_type_with_data (url, &data, parser);
+		DEBUG(g_print ("_get_mime_type_with_data for '%s' returned '%s'\n", url, mimetype ? mimetype : "NULL"));
 	}
 
 	if (mimetype == NULL)
@@ -1626,11 +1907,42 @@ totem_pl_parser_parse_internal (TotemPlParser *parser, const char *url)
 		return TOTEM_PL_PARSER_RESULT_IGNORED;
 	}
 
+	if (parser->priv->recurse || parser->priv->recurse_level == 0) {
+		parser->priv->recurse_level++;
+
+		for (i = 0; i < G_N_ELEMENTS(special_types); i++) {
+			if (strcmp (special_types[i].mimetype, mimetype) == 0) {
+				DEBUG(g_print ("URL '%s' is special type '%s'\n", url, mimetype));
+				ret = (* special_types[i].func) (parser, url, data);
+				found = TRUE;
+				break;
+			}
+		}
+
+		for (i = 0; i < G_N_ELEMENTS(dual_types) && found == FALSE; i++) {
+			if (strcmp (dual_types[i].mimetype, mimetype) == 0) {
+				DEBUG(g_print ("URL '%s' is dual type '%s'\n", url, mimetype));
+				if (data == NULL) {
+					mimetype = my_gnome_vfs_get_mime_type_with_data (url, &data, parser);
+				}
+				ret = (* dual_types[i].func) (parser, url, data);
+				found = TRUE;
+				break;
+			}
+		}
+
+		g_free (data);
+
+		parser->priv->recurse_level--;
+	}
+
+	if (ret == TOTEM_PL_PARSER_RESULT_SUCCESS)
+		return ret;
+
 	super = gnome_vfs_get_supertype_from_mime_type (mimetype);
-	for (i = 0; i < G_N_ELEMENTS (ignore_types); i++) {
+	for (i = 0; i < G_N_ELEMENTS (ignore_types) && super != NULL && found == FALSE; i++) {
 		if (gnome_vfs_mime_type_is_supertype (ignore_types[i].mimetype) != FALSE) {
 			if (strcmp (super, ignore_types[i].mimetype) == 0) {
-				g_free (data);
 				g_free (super);
 				return TOTEM_PL_PARSER_RESULT_IGNORED;
 			}
@@ -1639,34 +1951,12 @@ totem_pl_parser_parse_internal (TotemPlParser *parser, const char *url)
 
 			eq = gnome_vfs_mime_type_get_equivalence (mimetype, ignore_types[i].mimetype);
 			if (eq == GNOME_VFS_MIME_PARENT || eq == GNOME_VFS_MIME_IDENTICAL) {
-				g_free (data);
+				g_free (super);
 				return TOTEM_PL_PARSER_RESULT_IGNORED;
 			}
 		}
 	}
-
-	parser->priv->recurse_level++;
-
-	for (i = 0; i < G_N_ELEMENTS(special_types); i++) {
-		if (strcmp (special_types[i].mimetype, mimetype) == 0) {
-			ret = (* special_types[i].func) (parser, url, data);
-			g_free (data);
-			break;
-		}
-	}
-
-	for (i = 0; i < G_N_ELEMENTS(dual_types); i++) {
-		if (strcmp (dual_types[i].mimetype, mimetype) == 0) {
-			if (data == NULL) {
-				mimetype = my_gnome_vfs_get_mime_type_with_data (url, &data);
-			}
-			ret = (* dual_types[i].func) (parser, url, data);
-			g_free (data);
-			break;
-		}
-	}
-
-	parser->priv->recurse_level--;
+	g_free (super);
 
 	if (ret != TOTEM_PL_PARSER_RESULT_SUCCESS && parser->priv->fallback) {
 		totem_pl_parser_add_one_url (parser, url, NULL);
